 
/** @file iop_comedi.cpp Low-level access to physical device over comedi-driver*/

/*
   FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008, Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

*/


// include header
#include "iop_comedi.h"

namespace faudes {

// only compile for comedi support
#ifdef FAUDES_IODEVICE_COMEDI

// std faudes, incl dummy
FAUDES_TYPE_IMPLEMENTATION(ComediDevice,cDevice,sDevice)

// autoregister
AutoRegisterType<cDevice> gRtiRegisterComediDevice("ComediDevice");

//constructor
cDevice::cDevice(void) : sDevice(), mDev(0) {
  FD_DHV("cDevice(" << mName << ")::cDevice()");
  // have appropriate default label for token io
  mDefaultLabel = "ComediDevice";
}

//deconstructor
cDevice::~cDevice(void) {
  FD_DHV("cDevice(" << mName << ")::~cDevice()");
  Stop();
}

// Start(void)
void cDevice::Start(void) {
  //open comedi-device
  
  if(mState!=Down) return;
  FD_DH("cDevice(" << mName << ")::Start(): open devices");
  mDev=comedi_open(mSystemFile.c_str());
  // throw exception if opening device failed
  if(!mDev) {
    std::stringstream errstr;
    errstr << "cannot open device /dev/comedi0 (inputs)";
    throw Exception("cDevice()::Start()", errstr.str(), 552);
  }

  // check some properties on input card
  /*
  int sub, chan, dir;
  for(sub=0; sub<2; sub++) {
    for(chan=0; chan<32; chan++) {
      if(comedi_dio_get_config(DevIn, sub, chan, &dir) !=0) {
        FD_DH("dio::DioInitCards: error: on subdevice " << sub << " channel " << chan
          << "  /dev/comedi0 (inputs)");
        return 1;
      }
      if(dir != COMEDI_INPUT) {
        FD_DH("dio::DioInitCards: error: on subdevice " << sub << " channel " << chan
          << "  /dev/comedi0 (inputs)");
        return 1;
      }
    }
  }
  */

  // check some properties on output card
  /*
  for(sub=0; sub<2; sub++) {
    for(chan=0; chan<32; chan++) {
      if(comedi_dio_get_config(DevIn, sub, chan, &dir) !=0) {
        FD_DH("dio::DioInitCards: error: on subdevice " << sub << " channel " << chan
          << "  /dev/comedi1 (outputs)");
        return 1;
      }
      if(dir != COMEDI_INPUT) {
        FD_DH("dio::DioInitCards: error: on subdevice " << sub << " channel " << chan
          << "  /dev/comedi1 (outputs)");
        return 1;
      }
    }
  }
  */
  // call base
  sDevice::Start();
}//end Start()

// Stop()
void cDevice::Stop(void) {
  //close comedi-device

  if(mState != Up && mState != StartUp) return;
  FD_DHV("cDevice(" << mName << ")::Stop()");
  // call base
  sDevice::Stop();
  // close device
  comedi_close(mDev);
  mDev=0;
}

//DoWrite(rTr,rLabel,pContext)
void cDevice::DoWritePreface(TokenWriter& rTw, const std::string& rLabel,  const Type* pContext) const {
  FD_DHV("cDevice("<<mName<<")::DoWritePreface()");
  //call base
  sDevice::DoWritePreface(rTw,"",pContext);
  // write mSystemFile
  rTw.WriteString(mSystemFile);
}


//DoReadPreface(rTr,rLabel,pContext)
void cDevice::DoReadPreface(TokenReader& rTr,const std::string& rLabel, const Type* pContext){
  FD_DHV("cDevice("<<mName<<")::DoReadPreface()");
  //call base
  sDevice::DoReadPreface(rTr,"",pContext);


  // sense and digest pre 2.16 format
  Token token;
  rTr.Peek(token);
  if(token.IsString()) {
    mSystemFile = rTr.ReadString();
    return;
  }
  
  // loop my members
  while(true) {
    Token token;
    rTr.Peek(token);
    // system file
    if(token.IsBegin()) 
    if(token.StringValue()=="DeviceFile") {
      rTr.ReadBegin("DeviceFile", token);
      mSystemFile=token.AttributeStringValue("value");
      rTr.ReadEnd("DeviceFile");
      continue;
    }
    // break on unknown
    break;
  }

}

//ReadSignal(int)
bool cDevice::DoReadSignal(int bit){
  // read one input value, addressed by bit number (0 to 63);

  // todo: report errors, check range
  lsampl_t input=0;
  comedi_data_read(mDev,bit/32,bit%32,0,AREF_GROUND,&input);

  return (input!=0);
}


//WriteSignal(int,int)
void cDevice::DoWriteSignal(int bit, bool value){
  // write one actor value, adressed by bit number (0 to 63);

  // todo: report errors, check range
  lsampl_t output= (value ? 1 : 0);
  //std::cout << "dio::write " << bit << ": " << output << std::endl;
  comedi_data_write(mDev,bit/32,bit%32,0,AREF_GROUND,output);
}


#endif // if comedi

} // namespace


