/** @file syn_supnorm.cpp Supremal normal sublanguage */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

#include "syn_supnorm.h"
#include "syn_supcon.h"


namespace faudes {




/*
***************************************************************************************
***************************************************************************************
 Implementation
***************************************************************************************
***************************************************************************************
*/


//void NormalityConsistencyCheck(rL,rOAlph,rK) 
void NormalityConsistencyCheck(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK) {
  
  FD_DF("NormalityConsistencyCheck(...)");
  
  if (!(rK.IsDeterministic())) {
	 std::stringstream errstr;
     errstr << "Nondeterministic parameter rK.";
	 if(!(rL.IsDeterministic())) errstr << "Nondeterministic parameter rL.";
     throw Exception("NormalityConsistencyCheck", errstr.str(), 101);
	}
  if (!(rL.IsDeterministic())) {
	 std::stringstream errstr;
	 errstr << "Nondeterministic parameter rL.";
	 if(!(rL.IsDeterministic())) errstr << "Nondeterministic parameter rL.";
	 throw Exception("NormalityConsistencyCheck", errstr.str(), 101);
    }
  
  EventSet Kevents,Sigma;
  Sigma=rL.Alphabet();
  Kevents=rK.Alphabet();
  
  // observable events have to be subset of Sigma
  if(!(rOAlph<=Sigma)) {
    EventSet only_in_OAlph = rOAlph - Sigma;
    std::stringstream errstr;
    errstr << "Not all observable events are contained in Sigma: " 
       << only_in_OAlph.ToString() << ".";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 100);
  } 
	 
  // alphabets must match  
  if ( Sigma != Kevents) {
    EventSet only_in_L = Sigma - Kevents;
    EventSet only_in_K = Kevents - Sigma;
    only_in_L.Name("Only_In_L");
    only_in_L.Name("Only_In_K");
    std::stringstream errstr;
    errstr << "Alphabets of generators do not match.";
    if(!only_in_L.Empty())
      errstr <<  " " << only_in_L.ToString() << ".";
    if(!only_in_K.Empty())
      errstr <<  " " << only_in_K.ToString() << ".";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 100);
  }
  
  // K must be subset of L
  if (!LanguageInclusion(rK,rL)) {
    std::stringstream errstr;
    errstr << "K is not subset of L.";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 0);
  }
}

// IsNormal(rK,rL,rOAlph)
bool IsNormal(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK) 
{
  FD_DF("IsNormal(...)");

  // bail out on empty K
  // note: this is required to survive the
  // determinism test when rK has no states at all
  if(IsEmptyLanguage(rK)) return true;
  
  // determinism required
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  //extract overall alphabet:
  EventSet Sigma;
  Sigma=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents;
  Kevents=rK.Alphabet();
  
  //calculate p(K)
  Generator Ktest1, Ktest2;
  Ktest1.StateNamesEnabled(false);
  Ktest2.StateNamesEnabled(false);
  Project(rK,rOAlph,Ktest1);
  
  //calculate pinv(p(K))
  InvProject(Ktest1, Sigma);
  
  //check normality: pinv(p(K)) intersect L = K?
  LanguageIntersection(Ktest1,rL,Ktest2);
  return LanguageInclusion(Ktest2,rK);
	
  // Remark: testing for LanguageEquality is not required, as inclusion
  // in the reverse direction is always met (assuming K\subseteq L)
}



// IsNormal rti wrapper
bool IsNormal(const System& rPlantGen, const Generator& rSupCandGen) {
  return IsNormal(rPlantGen, rPlantGen.ObservableEvents(),rSupCandGen);
}


// ConcatenateFullLanguage(rGen)
void ConcatenateFullLanguage(Generator& rGen) {
  FD_DF("ConcatenateFullLanguage(" << rGen.Name() << ")");
  
  // prepare result
  rGen.Name("ConcatenateFullLanguage(" + rGen.Name() + ")");
  
  // treat trivial empty result in case of empty marked language
  if(rGen.MarkedStatesSize()==0) {
		rGen.Clear();
		return;
	}
  
  // helpers
  EventSet alph=rGen.Alphabet();
  StateSet StatesToClear;
  TransSet TransToClear,TransToSet;
  StateSet::Iterator sit;
  TransSet::Iterator tit;
  EventSet::Iterator evit;
  
  // clear all transitions in marked states
  for (sit = rGen.MarkedStatesBegin(); sit != rGen.MarkedStatesEnd(); ++sit) {
      for (tit = rGen.TransRelBegin(*sit); tit != rGen.TransRelEnd(*sit); ++tit) {
          TransToClear.Insert(*tit);
      }
  }
  
  for (tit = TransToClear.Begin(); tit != TransToClear.End(); ++tit) {
      rGen.ClrTransition(*tit);
  }
  TransToClear.Clear();
  rGen.Accessible();
  
  // all marked states become eqivalent -> switch transitions leading to marked states to one remaining marked state
  // and delete all the others; note: for this special purpose, this procedure is faster than running StateMin.
  // if an init state is marked the effort is even less (as the result is Sigma* itself, see "else" below)
  bool initmarked = !( (rGen.InitStates() * rGen.MarkedStates()).Empty() );
  if (!initmarked) {
	  // define remaining marked state (there is one because of nonzero MarkedStatesSize)
	  sit = rGen.MarkedStatesBegin();
	  Idx marked = *sit;
	  // switch transitions to all other marked states if there are any
	  if(rGen.MarkedStatesSize()>1) {
		  // extract transitions sorted by target state X2
		  TransSetX2EvX1 trel;
		  rGen.TransRel(trel);
		  TransSetX2EvX1::Iterator tit2;
		  
		  // switch transitions to all other marked states to the one remaining marked state
		  ++sit;
		  for (; sit != rGen.MarkedStatesEnd(); ++sit) {
		      for (tit2 = trel.BeginByX2(*sit); tit2 != trel.EndByX2(*sit); ++tit2) {
				  // switch transition target to remaining marked state
				  TransToSet.Insert(tit2->X1,tit2->Ev,marked);
				  // note: original transitions are cleared later with clearing the corresp. marked states.
		      }
		  }
		  
		  // delete all but remaining marked state (note: by doing so, also corresp. transitions are cleared.)
		  sit = rGen.MarkedStatesBegin();
		  ++sit;
		  for (; sit != rGen.MarkedStatesEnd(); ++sit) {
			StatesToClear.Insert(*sit);
		  }
		  for (sit=StatesToClear.Begin(); sit != StatesToClear.End(); ++sit) {
			rGen.DelState(*sit);
		  }
		  StatesToClear.Clear();
		  
		  // insert switched transitions
		  for (tit = TransToSet.Begin(); tit != TransToSet.End(); ++tit) {
		      rGen.SetTransition(*tit);
		  }
		  TransToSet.Clear();
	    }

	  // now concatenate Sigma* by selflooping marked state with all events
	  for (evit = alph.Begin(); evit != alph.End(); ++evit) {
		  rGen.SetTransition(marked,*evit,marked);
	  }
	  
	}
  // consider marked init state case
  else {
		// reduce to empty string language
		rGen.ClearStates();
		rGen.ClearTransRel();
		Idx state=rGen.InsInitState();
		rGen.SetMarkedState(state);
		// now concatenate Sigma* by selflooping marked state with all events
		for (evit = alph.Begin(); evit != alph.End(); ++evit) {
		  rGen.SetTransition(state,*evit,state);
		}
		
    }
  
  return;
}


//  SupNorm(rL,rOAlph,rK,rResult)
bool SupNorm(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupNorm(" << rL.Name() << "," << rK.Name() << "," << rOAlph.Name() << ")");
  
  // exceprions
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  //extract overall alphabet:
  EventSet allevents;
  allevents=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents=rK.Alphabet();
    
  // involved operations from cfl_regular.h operate on the marked
  // languages -> turn generated languages into marked langusges
  Generator prL=rL;
  prL.InjectMarkedStates(rL.States());
  Generator prK=rK;
  prK.InjectMarkedStates(prK.States());
  
  // calculate "L-K" 
  LanguageDifference(prL,prK,rResult);
  
  // calculate Pinv(P(L-K)):
  Project(rResult,rOAlph,rResult);
  FD_DF("SupNorm: sizeof p(L-K): "<< rResult.Size());
  InvProject(rResult,allevents);
  FD_DF("SupNorm: sizeof pinv(p(L-K)): "<< rResult.Size());

  //calculate remaining set difference -> supnorm(K)
  LanguageDifference(prK,rResult,rResult);
  
  // cosmetics: remove blocking states
  rResult.Trim();
   
  // done
  rResult.Name("SupNorm("+rL.Name()+", "+rK.Name()+")");  
  return !( rResult.InitStatesEmpty() );
}

//  SupNormClosed(rL,rOAlph,rK,rResult)
bool SupNormClosed(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupNormClosed(" << rL.Name() << "," << rK.Name() << "," << rOAlph.Name() << ")");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  //extract overall alphabet:
  EventSet allevents;
  allevents=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents=rK.Alphabet();
  
  // involved operations from regular.h operate on the marked
  // languages -> turn generated languages into marked langs
  Generator prL=rL;
  prL.InjectMarkedStates(prL.States());
  Generator prK=rK;
  prK.InjectMarkedStates(prK.States());
  
  // calculate "L-K" 
  LanguageDifference(prL,prK,rResult);

  // calculate Pinv(P(L-K)):
  Project(rResult,rOAlph,rResult);
  FD_DF("SupNormClosed: sizeof p(L-K): "<< rResult.Size());
  InvProject(rResult,allevents);
  FD_DF("SupNormClosed: sizeof pinv(p(L-K)): "<< rResult.Size());
  
  //concatenate Pinv(P(L-K)) with Sigma*: this leads to closed result
  ConcatenateFullLanguage(rResult); // special function for efficient concatenation of Sigma*.
  
  FD_DF("SupNormClosed: sizeof pinv(p(L-K))Sigma*: "<< rResult.Size());

  //calculate remaining set difference -> supnormClosed(K)
  LanguageDifference(prK,rResult,rResult);

  // cosmetics: remove blocking states
  rResult.Trim();

  // done
  rResult.Name("SupNormClosed("+rL.Name()+", "+rK.Name()+")"); 
  return !( rResult.InitStatesEmpty() );
}


//  SupConNorm(rL,rCAlph,rOAlph,rK,rResult)
void SupConNorm(
  const Generator& rL,
  const EventSet& rCAlph,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupConNorm(" << rL.Name() << "," << rK.Name() << ")");
  // 0.: intersect K with L to match requirements of SupNormClosed
  Generator K;
  K.StateNamesEnabled(false);
  Product(rL,rK,K);
  // 1. normal and closed sublanguage (operates on / returns generated language)
  Generator N;
  N.StateNamesEnabled(false);
  SupNormClosed(rL,rOAlph,K,N);
  // 2. project to sigma_o (generated languages)
  Generator N0,L0;
  N0.StateNamesEnabled(false);
  L0.StateNamesEnabled(false);
  Project(N,rOAlph,N0);
  Project(rL,rOAlph,L0);
  // 3. supremal controllable sublanguage (generated languages)
  EventSet sig_co = rCAlph * rOAlph;
  Generator K0;
  K0.StateNamesEnabled(false);
  SupCon(L0,sig_co,N0,K0);
  // 4. inverese project to sigma  (on generated language)
  InvProject(K0,rL.Alphabet());
  // 5. intersect with L (generated languages)
  LanguageIntersection(K0,rL,rResult);
  // convenience: mark the generated language
  rResult.InjectMarkedStates(rResult.States());
}  


//  SupConNormNB(rL,rCAlph,rOAlph,rK,rResult)
void SupConNormNB(
  const Generator& rL,
  const EventSet& rCAlph,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupConNormNB(" << rL.Name() << "," << rK.Name() << ")");
  // initialize: K0
  Generator K0;
  K0.StateNamesEnabled(false);
  Product(rL,rK,K0);
  K0.Coaccessible();
  // initialize: closure(rL)
  Generator L=rL;
  L.StateNamesEnabled(false);
  L.Trim();
  L.InjectMarkedStates(L.States());
  // loop
  Generator Ki=K0;
  Ki.StateNamesEnabled(false);
  while(1) {
    FD_DF("SupConNormNB(" << rL.Name() << "," << rK.Name() << "): #" << Ki.Size() << " m#" << Ki.MarkedStatesSize());
    // keep copy of recent
    rResult=Ki;
    // cheep closure (for coreachable generator)
    Ki.InjectMarkedStates(Ki.States());
    // synthesise closed
    SupConNorm(L,rCAlph,rOAlph,Ki,Ki);
    // restrict
    Product(K0,Ki,Ki);
    Ki.Coaccessible();
    // test (sequence is decreasing anyway)
    if(LanguageInclusion(rResult,Ki)) break;
  }  
  FD_DF("SupConNormNB(" << rL.Name() << "," << rK.Name() << "): done");
}  


// SupPrefixClosed(rK,rResult)
bool SupPrefixClosed(
  const Generator& rK,
	Generator& rResult) {
  
	FD_DF("SupPrefixClosed("<<rK.Name()<<")");
  
  // prepare Result:
  rResult.Name("SupPrefixClosed("+rK.Name()+")");
  
  // check for marked initial state, empty result if not
  if( (rK.InitStates() * rK.MarkedStates()).Empty() ) {
    rResult.Clear();
    return false;
  }

  rResult.Assign(rK);  
  
  // erase all transitions not leading to a marked state
  // todo: Depth-first-search could be faster
  TransSet trel=rResult.TransRel();
  TransSet::Iterator tit=trel.Begin();
  TransSet::Iterator tit_end=trel.End();
  for(;tit!=tit_end;tit++) {
    if(!(rResult.ExistsMarkedState(tit->X2))) {
		  rResult.ClrTransition(*tit);
			}
  }
  
  // make reachable (cosmetic)
  rResult.Trim();
  
  // as there is at least one marked init state, result is nonempty
  return true;
}



/** rti wrapper */
void SupNorm(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{
  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  // execute
  SupNorm(rPlantGen,rPlantGen.ObservableEvents(),rSpecGen,*pResGen);
  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}

/** rti wrapper */
void SupNormClosed(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{
  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  // execute
  SupNormClosed(rPlantGen,rPlantGen.ObservableEvents(),rSpecGen,*pResGen);
  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}


/** rti wrapper */
void SupConNorm(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{
  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  // execute
  SupConNorm(rPlantGen,rPlantGen.ControllableEvents(),rPlantGen.ObservableEvents(),rSpecGen,*pResGen);
  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}

/** rti wrapper */
void SupConNormNB(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{
  FD_WARN("SupConNormNB(" << rPlantGen.Name() << "," << rSpecGen.Name() << "): rti wrapper");
  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  // execute
  SupConNormNB(rPlantGen,rPlantGen.ControllableEvents(),rPlantGen.ObservableEvents(),rSpecGen,*pResGen);
  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}

} // end namespace
