/** @file syn_wsupcon.cpp Supremal controllable sublanguage for infinite time behaviours */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2010 Thomas Moor

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

//#define FAUDES_DEBUG_FUNCTION

#include "syn_wsupcon.h"
#include "syn_supcon.h"
#include "syn_supnorm.h"
#include "syn_functions.h"


namespace faudes {


/*
***************************************************************************************
***************************************************************************************
 Implementation IsOmegaControllabe
***************************************************************************************
***************************************************************************************
*/

// IsOmegaControllable()
bool IsOmegaControllable(
  const Generator& rGenPlant, 
  const EventSet&  rCAlph, 
  const Generator& rGenCand) 
{
  FD_DF("IsOmegaControllable(\"" <<  rGenPlant.Name() << "\", \"" << rGenCand.Name() << "\")");

  // alphabets must match
  if ( rGenPlant.Alphabet() != rGenCand.Alphabet()) {
    std::stringstream errstr;
    errstr << "Alphabets of generators do not match.";
    throw Exception("IsOmegaControllable(..)", errstr.str(), 100);
  }

#ifdef FAUDES_CHECKED
  // generators are meant to be nonblocking
  if( !IsOmegaTrim(rGenCand) ) {
    std::stringstream errstr;
    errstr << "Argument \"" << rGenCand.Name() << "\" is not omega-trim.";
    throw Exception("IsOmegaControllable(..)", errstr.str(), 201);
  }
  if( !IsOmegaTrim(rGenPlant) ) {
    std::stringstream errstr;
    errstr << "Argument \"" << rGenPlant.Name() << "\" is not omega-trim.";
    throw Exception("IsOmegaControllable(..)", errstr.str(), 201);
  }
#endif

  // the trivial case: empty cand is fine
  // (we must treat this case because empty generators are not regarded deterministic)
  if(rGenCand.Empty()) {
    FD_DF("IsOmegaControllable(..): empty candidate, pass");
    return true;
  }

  // the trivial case: empty plant is fails
  // (we must treat this case because empty generators are not regarded deterministic)
  if(rGenPlant.Empty()) {
    FD_DF("IsOmegaControllable(..): empty plant, fail");
    return false;
  }

#ifdef FAUDES_CHECKED
  // generators are meant to be deterministic
  if ( !IsDeterministic(rGenCand) ||  !IsDeterministic(rGenPlant)) {
    std::stringstream errstr;
    errstr << "Arguments are expected to be deterministic.";
    throw Exception("IsRelativelyOmegaClosed", errstr.str(), 202);
  }
#endif

  // test controllability
  StateSet dummy;
  if(!IsControllableUnchecked(rGenPlant,rCAlph,rGenCand,dummy)) return false;

  // test relative closedness
  if(!IsRelativelyOmegaClosedUnchecked(rGenPlant,rGenCand)) return false;

  // pass
  FD_DF("IsOmegaControllable(...): passed");
  return true;
}


// IsOmegaControllable() wrapper
bool IsOmegaControllable(
  const System& rPlantGen, 
  const Generator& rCandGen) 
{
  return IsOmegaControllable(rPlantGen,rPlantGen.ControllableEvents(),rCandGen);
}

/*
***************************************************************************************
***************************************************************************************
 Implementation SupConCmpl
***************************************************************************************
***************************************************************************************
*/


// supcon complete
void SupConCmpl(
  const Generator& rPlantGen, 
  const EventSet&  rCAlph,
  const Generator& rSpecGen, 
  Generator& rResGen)
{ 
  FD_DF("SupConCmpl(" << rPlantGen.Name() << "," << rSpecGen.Name()<< ")");
  
  // exceptions on invalid parameters, same as std synthesis
  ControlProblemConsistencyCheck(rPlantGen,rCAlph,rSpecGen);  

  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name("SupConCmpl("+rPlantGen.Name()+", "+rSpecGen.Name()+")");
  pResGen->InjectAlphabet(rPlantGen.Alphabet());
  
  //check for trivial result
  if(rSpecGen.InitStatesEmpty()){
    FD_DF("SupConCmpl: empty language specification - empty result.");
  }  
    

  // have a reverse composition map
  std::map< std::pair<Idx,Idx>, Idx> revmap;

  // parallel composition (result is reachable)
  SupconParallel(rPlantGen, rCAlph, rSpecGen, revmap, *pResGen);

  // make resulting generator complete and controllabel
  while(true) {
    Idx state_num = pResGen->Size();
    if(pResGen->InitStates().Empty()) break;
    pResGen->Complete();
    if(pResGen->InitStates().Empty()) break;
    SupConUnchecked(rPlantGen, rCAlph, *pResGen);
    if(pResGen->Size() == state_num) break;
  }

  // convenience state names
  if(rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && rResGen.StateNamesEnabled()) 
    SetComposedStateNames(rPlantGen, rSpecGen, revmap, *pResGen);
  else
    pResGen->StateNamesEnabled(false);

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}


// user wrapper
void SupConCmpl(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{

  // execute 
  SupConCmpl(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,rResGen);

  // copy all attributes of input alphabet
  rResGen.EventAttributes(rPlantGen.Alphabet());

}




/*
***************************************************************************************
***************************************************************************************
 Implementation SupConCmplNB
***************************************************************************************
***************************************************************************************
*/


// supcon complete
void SupConCmplNB(
  const Generator& rPlantGen, 
  const EventSet&  rCAlph,
  const Generator& rSpecGen, 
  Generator& rResGen)
{ 
  FD_DF("SupConCmplNB(" << rPlantGen.Name() << "," << rSpecGen.Name()<< ")");
  
  // exceptions on invalid parameters, same as std synthesis
  ControlProblemConsistencyCheck(rPlantGen,rCAlph,rSpecGen);  

  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name("SupConCmplNB("+rPlantGen.Name()+", "+rSpecGen.Name()+")");
  pResGen->InjectAlphabet(rPlantGen.Alphabet());
  
  //check for trivial result
  if(rSpecGen.InitStatesEmpty()){
    FD_DF("SupConCmplNB: empty language specification - empty result.");
  }  
    

  // have a reverse composition map
  std::map< std::pair<Idx,Idx>, Idx> revmap;

  // parallel composition (result is reachable)
  SupconParallel(rPlantGen, rCAlph, rSpecGen, revmap, *pResGen);

  // make resulting generator complete and controllabel and coaccessible
  while(true) {
    Idx state_num = pResGen->Size();
    if(pResGen->InitStates().Empty()) break;
    pResGen->Coaccessible();
    if(pResGen->InitStates().Empty()) break;
    pResGen->Complete();
    if(pResGen->InitStates().Empty()) break;
    SupConUnchecked(rPlantGen, rCAlph, *pResGen);
    if(pResGen->Size() == state_num) break;
  }

  // convenience state names
  if(rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && rResGen.StateNamesEnabled()) 
    SetComposedStateNames(rPlantGen, rSpecGen, revmap, *pResGen);
  else
    pResGen->StateNamesEnabled(false);

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}


// user wrapper
void SupConCmplNB(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{

  // execute 
  SupConCmplNB(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,rResGen);

  // copy all attributes of input alphabet
  rResGen.EventAttributes(rPlantGen.Alphabet());

}



/*
***************************************************************************************
***************************************************************************************
 Implementation SupConNormCmplNB
***************************************************************************************
***************************************************************************************
*/


//  SupConNormCmplNB(rL,rCAlph,rOAlph,rK,rResult)
void SupConNormCmplNB(
  const Generator& rL,
  const EventSet& rCAlph,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupConNormCmplNB(" << rL.Name() << "," << rK.Name() << ")");
  // initialize: K0
  Generator K0;
  K0.StateNamesEnabled(false);
  Product(rL,rK,K0);
  K0.Coaccessible();
  // initialize: closure(rL)
  Generator L=rL;
  L.StateNamesEnabled(false);
  L.Trim();
  L.InjectMarkedStates(L.States());
  // loop
  Generator Ki=K0;
  Ki.StateNamesEnabled(false);
  while(1) {
    FD_DF("SupConNormCmplNB(" << rL.Name() << "," << rK.Name() << "): #" << Ki.Size() << " m#" << Ki.MarkedStatesSize());
    // keep copy of recent
    rResult=Ki;
    // cheep closure (for coreachable generator)
    Ki.InjectMarkedStates(Ki.States());
    // synthesise closed
    SupConNorm(L,rCAlph,rOAlph,Ki,Ki);
    Complete(Ki);
    // restrict
    Product(K0,Ki,Ki);
    Ki.Coaccessible();
    // test (sequence is decreasing anyway)
    if(LanguageInclusion(rResult,Ki)) break;
  }  
  FD_DF("SupConNormCmplNB(" << rL.Name() << "," << rK.Name() << "): done");
}  


/** rti wrapper */
void SupConNormCmplNB(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{
  FD_DF("SupConNormCmplNB(" << rPlantGen.Name() << "," << rSpecGen.Name() << "): rti wrapper");
  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  // execute
  SupConNormCmplNB(rPlantGen,rPlantGen.ControllableEvents(),rPlantGen.ObservableEvents(),rSpecGen,*pResGen);
  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());
  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}


/*
***************************************************************************************
***************************************************************************************
 Implementation OmegaSupConNB(experimental)
***************************************************************************************
***************************************************************************************
*/





// OmegaControlledLiveness(...)
bool OmegaControlledLiveness(
  Generator& rSupCandGen, 
  const EventSet& rCAlph,
  StateSet& rMarkedPlantStates, 
  StateSet& rMarkedSpecStates)
{
  FD_DF("OmegaControlledLiveness(...)");


  // set of critical states
  StateSet critical;
   
  // return true if parallelcomp contains no initial states
  if(rSupCandGen.InitStatesEmpty()) {
    return true;
  }

#ifdef FAUDES_DEBUG_FUNCTION
  FD_DF("OmegaControlledLiveness(): marked states: ");
  StateSet ssd= rMarkedPlantStates + rMarkedSpecStates;
  ssd.Write();
#endif


  // find un-marked sccs
  StateSet astates = rMarkedPlantStates + rMarkedSpecStates;
  SccFilter umfilter(SccFilter::FmIgnoreTrivial | SccFilter::FmStatesAvoid, astates);
  std::list<StateSet> umsccs;
  StateSet umroots;
  ComputeScc(rSupCandGen,umfilter,umsccs,umroots); 

  // report
  std::list<StateSet>::iterator ssit=umsccs.begin();
  for(;ssit!=umsccs.end(); ++ssit) {
    FD_DF("OmegaControlledLiveness(): unmarked scc: " << ssit->ToString());
  }  


  // good-states iteration
  StateSet goodstates = rMarkedSpecStates;
  

  // drive states to good states
  while(true) {
    // LoopCallback();
    // allow for user interrupt, incl progress report
    FD_WPC(1,2,"OmegaControlledLiveness(): iterating states"); 
    // test individual states
    FD_DF("OmegaControlledLiveness(): iterate over states (#" << rSupCandGen.Size() << ")");
    bool found=false;

    StateSet::Iterator sit = rSupCandGen.StatesBegin();
    StateSet::Iterator sit_end = rSupCandGen.StatesEnd();
    for(; sit!=sit_end; ++sit) {
      bool fail = false;  // cbaier 20121011
      bool positive=false; // cbaier 20121011
      // goodstate anyway
      if(goodstates.Exists(*sit)) continue;
      // test transitions
      TransSet::Iterator tit = rSupCandGen.TransRelBegin(*sit);
      TransSet::Iterator tit_end = rSupCandGen.TransRelEnd(*sit);
      // no transitions at all
      if(tit==tit_end) continue;
      // loop over successors
      for(; tit!=tit_end; ++tit) {
        if(goodstates.Exists(tit->X2)) 
	{positive=true; // cbaier 20121011: added 
	continue;}
        if(rCAlph.Exists(tit->Ev)) continue;
        
      // good states survive the loop // cbaier 20121011: added
      if(tit!=tit_end) { 
	fail=true;
        break;
        }                            
      }
      
      // good states survive
      if(!fail && positive) {    // cbaier 20121011: changed
        FD_DF("OmegaControlledLiveness(): good state " << rSupCandGen.SStr(*sit));
        goodstates.Insert(*sit);
        found=true;
      }
    }  
      
    // test individual unmarked sccs
    FD_DF("OmegaControlledLiveness(): iterate over unmarked sccs (#" << umsccs.size() <<")");
    std::list<StateSet>::iterator ssit=umsccs.begin();
    while(ssit!=umsccs.end()) { 
      bool fail=false;
      bool positive=false;
      sit=ssit->Begin();
      sit_end=ssit->End();
      for(; sit!=sit_end; ++sit) {
        // goodstate anyway
        if(goodstates.Exists(*sit)) continue;
        // test transitions
        TransSet::Iterator tit = rSupCandGen.TransRelBegin(*sit);
        TransSet::Iterator tit_end = rSupCandGen.TransRelEnd(*sit);
        // no transitions at all
        if(tit==tit_end) continue;        
        // loop over successors
        for(; tit!=tit_end; ++tit) {
          if(goodstates.Exists(tit->X2)) { positive=true; continue;}
          if(rCAlph.Exists(tit->Ev)) continue;  // tmoor 20110202: fixed typo
          if(ssit->Exists(tit->X2)) continue;
          break;
        }
        // good states survive the loop
        if(tit!=tit_end) {
          fail=true;
          break;
        }
      }
      
      // prepare next scc iterator
      std::list<StateSet>::iterator ssitnext=ssit;
      ++ssitnext;
      bool lend= (ssitnext==umsccs.end());
      // all states passed, then they are all good
      if(!fail && positive) {
        FD_DF("OmegaControlledLiveness(): good scc " << ssit->ToString());
        goodstates.InsertSet(*ssit);
        umsccs.erase(ssit);
        found=true;
      }
      // break on end of list
      if(lend) break;
      // progress
      ssit=ssitnext;
    }
    
    // exit
    if(!found) break;
  };


  // delete critical states
  FD_DF("OmegaControlledLiveness(): good states: " << goodstates.ToString())
  StateSet::Iterator sit = rSupCandGen.StatesBegin();
  StateSet::Iterator sit_end = rSupCandGen.StatesEnd();
  for(; sit!=sit_end; ++sit) {
    if(!goodstates.Exists(*sit))
      rSupCandGen.DelState(*sit);
  }

  // return true
  FD_DF("OmegaControlledLiveness(): done");
  return true;
}


// helper class
class OPSState {
public:
  // minimal interface
  OPSState() {};
  OPSState(const Idx& rq1, const Idx& rq2, const bool& rf) :
    q1(rq1), q2(rq2), m1required(rf) {};
  std::string Str(void) { return ToStringInteger(q1)+"|"+
      ToStringInteger(q2)+"|"+ToStringInteger(m1required); };
  // order
  bool operator < (const OPSState& other) const {
    if (q1 < other.q1) return(true);
    if (q1 > other.q1) return(false);
    if (q2 < other.q2) return(true);
    if (q2 > other.q2) return(false);
    if (m1required && !other.m1required) return(true);
    return(false);
  }
  // member variables
  Idx q1;
  Idx q2;
  bool m1required;
};


// OmegaSupconParallel(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, rResGen)
void OmegaSupconParallel(
  const Generator& rPlantGen, 
  const EventSet& rCAlph,
  const Generator& rSpecGen,
  std::map< OPSState , Idx>& rReverseCompositionMap, 
  StateSet& rMarkedPlantStates, 
  StateSet& rMarkedSpecStates, 
  Generator& rResGen) 
{
  FD_DF("OmegaSupconParallel(" << &rPlantGen << "," << &rSpecGen << ")");
	
  // prepare result
  rResGen.Clear();
  rResGen.InjectAlphabet(rPlantGen.Alphabet());

  // trivial case
  if (rPlantGen.InitStatesEmpty()) {
    FD_DF("OmegaSupconParallel: plant got no initial states. "
            << "parallel composition contains empty language.");
    return;
  }
  if (rSpecGen.InitStatesEmpty()) {
    FD_DF("OmegaSupconParallel: spec got no initial states. "
              << "parallel composition contains empty language.");
    return;
  }

  // todo stack
  std::stack< OPSState > todo;
  // current pair, new pair
  OPSState currentstates, newstates;
  // state
  Idx tmpstate;  
  TransSet::Iterator ptit, ptit_end, stit, stit_end;
  std::map< OPSState, Idx>::iterator rcit;
  // critical states
  StateSet critical;


  // push initial state on todo stack
  currentstates = OPSState(*rPlantGen.InitStatesBegin(), *rSpecGen.InitStatesBegin(), true);
  todo.push(currentstates);
  tmpstate=rResGen.InsInitState();
  rReverseCompositionMap[currentstates] = tmpstate;
  if(rPlantGen.ExistsMarkedState(currentstates.q1)) 
    rMarkedPlantStates.Insert(tmpstate);
  if(rSpecGen.ExistsMarkedState(currentstates.q2)) 
    rMarkedSpecStates.Insert(tmpstate);

  // start algorithm
  FD_DF("OmegaSupconParallel: processing reachable states:");
  while(!todo.empty()) {
    // allow for user interrupt
    // LoopCallback();
    // allow for user interrupt, incl progress report
    FD_WPC(rReverseCompositionMap.size(),rReverseCompositionMap.size()+todo.size(),"OmegaSupconParallel(): processing"); 
    // get next reachable state from todo stack
    currentstates = todo.top();
    todo.pop();
    FD_DF("OmegaSupconParallel: processing (" << currentstates.Str() << " -> " 
	  << rReverseCompositionMap[currentstates]<<")");
    // iterate over all rPlantGen and rSpecGen transitions 
    ptit = rPlantGen.TransRelBegin(currentstates.q1);
    ptit_end = rPlantGen.TransRelEnd(currentstates.q1);
    stit = rSpecGen.TransRelBegin(currentstates.q2);
    stit_end = rSpecGen.TransRelEnd(currentstates.q2);
    // process all transitions and increment iterators strategically
    while((ptit != ptit_end) && (stit != stit_end)) {
      FD_DF("OmegaSupconParallel: current plant-transition: " << rPlantGen.TStr(*ptit) );
      FD_DF("OmegaSupconParallel: current spec-transition: " << rSpecGen.TStr(*stit));
      // increment tit1 and tit2, case A: execute common event
      if(ptit->Ev == stit->Ev) {
	FD_DF("OmegaSupconParallel: executing common event "  << rPlantGen.EStr(ptit->Ev));
        newstates = OPSState(ptit->X2, stit->X2,currentstates.m1required);
        // track marking requirement
        if(currentstates.m1required) {
  	  if(rPlantGen.ExistsMarkedState(currentstates.q1))
	    newstates.m1required=false;
        } else {
  	  if(rSpecGen.ExistsMarkedState(currentstates.q2)) { 
	    newstates.m1required=true;
          }
        }
	rcit = rReverseCompositionMap.find(newstates);
	// if state is new: add to todo list and result
	if(rcit == rReverseCompositionMap.end()) {
          todo.push(newstates);
          tmpstate = rResGen.InsState();
          rReverseCompositionMap[newstates] = tmpstate;
          if(!newstates.m1required) 
          if(rSpecGen.ExistsMarkedState(newstates.q2)) 
            rResGen.SetMarkedState(tmpstate);                   
          if(rPlantGen.ExistsMarkedState(newstates.q1)) 
            rMarkedPlantStates.Insert(tmpstate);
          if(rSpecGen.ExistsMarkedState(newstates.q2)) 
            rMarkedSpecStates.Insert(tmpstate);
          FD_DF("OmegaSupconParallel:   todo push: (" << newstates.Str() << ") -> " 
              << rReverseCompositionMap[newstates]);
        } else {
          tmpstate = rcit->second;
        }
	// if successor state is not critical add transition and increment iterators
	if(!critical.Exists(tmpstate)) {
          rResGen.SetTransition(rReverseCompositionMap[currentstates], 
              ptit->Ev, tmpstate);
          FD_DF("OmegaSuperParallel:  add transition to new generator: " << 
		rResGen.TStr(Transition(rReverseCompositionMap[currentstates], ptit->Ev, tmpstate)));
          ++ptit;
          ++stit;
        }
	// if successor state is critical and event is uncontrollable then delete current state 
	else if(!rCAlph.Exists(ptit->Ev)) {
	  FD_DF("OmegaSupconOmegaParallel: successor " << rSpecGen.SStr(tmpstate) 
		       << "in critical and common event " << rSpecGen.EStr(stit->Ev) 
		       << " uncontrollable:");
	  FD_DF("OmegaSupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	  critical.Insert(tmpstate);
#ifdef FAUDES_CHECKED
	  // do not end while loops here for detecting all h transitions 
	  // not in spec
	  FD_DF("OmegaSupconParallel: incrementing plant and spec transrel (FAUDES_CHECKED)");
	  ++ptit;
	  ++stit;
#else
	  // exit all loops
	  ptit = ptit_end;
	  stit = stit_end;
#endif
        }
        // else if successor state is critical and event controllable increment iterators
        else {
	  FD_DF("OmegaSupconParallel: incrementing plant and spec transrel");
          ++ptit;
          ++stit;
        }
      } // end: if processing common event 
      // increment iterators case B: process plant event that is not enebled for h
      else if (ptit->Ev < stit->Ev) {
	FD_DF("SupconParallel: asynchronous execution of event " 
		 << rPlantGen.EStr(ptit->Ev) << " in plant while " << rSpecGen.EStr(stit->Ev) << " in spec");
	// if uncontrollable transition leaves specification
	// delete state from res and put into criticalset
	if (!rCAlph.Exists(ptit->Ev)) {
	  FD_DF("OmegaSupconParallel: asynchronous event " << rPlantGen.EStr(ptit->Ev) 
	     << " in g is uncontrollable");
	  tmpstate = rReverseCompositionMap[currentstates];	
	  FD_DF("OmegaSupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	  critical.Insert(tmpstate);
	  // exit all loops
	  ptit = ptit_end;
	  stit = stit_end;
	  break;
	}
	FD_DF("OmegaSupconParallel: incrementing plant transrel");
	++ptit;
      } 
      // increment titg and titg, case C: process h event that is not enabled for g
      else {
	//#ifdef FAUDES_CHECKED
	// FD_WARN("SupconParallel:  transition " << rSpecGen.TStr(*tith) << " in specification h not found in g");
	//#endif
	FD_DF("OmegaSupconParallel: incrementing spec transrel");
	++stit;
      }
    } // end while incrementing 
    // increment iterators case D: process leftover events of plant
    while (ptit != ptit_end) {
      FD_DF("OmegaSupconParallel: asynchronous execution of event " 
	       << rPlantGen.EStr(ptit->Ev) << " in plant at end of spec");
      FD_DF("OmegaSupconParallel: curremt g-transition: " << rPlantGen.SStr(ptit->X1) 
	       << "-" << rPlantGen.EStr(ptit->Ev) << "-" << rPlantGen.SStr(ptit->X2));
      FD_DF("OmegaSupconParallel: current h-transition: end");
      // if uncontrollable transition leaves specification
      if (!rCAlph.Exists(ptit->Ev)) {
	tmpstate = rReverseCompositionMap[currentstates];
	FD_DF("OmegaSupconParallel: asynchron executed uncontrollable end "
	         << "event " << rPlantGen.EStr(ptit->Ev) << " leaves specification:");
	FD_DF("SupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	critical.Insert(tmpstate);
	// exit this loop 
	break;
      }
      FD_DF("OmegaSupconParallel: incrementing g transrel");
      ++ptit;
    }
  } // while todo
  FD_DF("OmegaSupconParallel: deleting critical states: " << critical.ToString());
  rResGen.DelStates(critical);
  rMarkedPlantStates.EraseSet(critical);
  rMarkedSpecStates.EraseSet(critical);

}



// OmegaSupConNBUnchecked(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, rResGen)
void OmegaSupConNBUnchecked(
  const Generator& rPlantGen,
  const EventSet& rCAlph,  
  const Generator& rSpecGen,
  std::map< OPSState , Idx>& rReverseCompositionMap, 
  Generator& rResGen) 
{
  FD_DF("OmegaSupConNB(\"" <<  rPlantGen.Name() << "\", \"" << rSpecGen.Name() << "\")");

  // PREPARE RESULT:	
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name(CollapsString("OmegaSupConNB(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))"));
  pResGen->InjectAlphabet(rPlantGen.Alphabet());

  // controllable events
  FD_DF("OmegaSupConNB: controllable events: "   << rCAlph.ToString());

  // ALGORITHM:
  StateSet plantmarking;
  StateSet specmarking;
  OmegaSupconParallel(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, plantmarking, specmarking, *pResGen);



  // fix statenames ...
  std::map< OPSState, Idx>::iterator rcit;
  for(rcit=rReverseCompositionMap.begin(); rcit!=rReverseCompositionMap.end(); rcit++) {
    Idx x1=rcit->first.q1;
    Idx x2=rcit->first.q2;
    bool m1requ=rcit->first.m1required;
    Idx x12=rcit->second;
    if(!pResGen->ExistsState(x12)) continue;
    std::string name1= rPlantGen.StateName(x1);
    if(name1=="") name1=ToStringInteger(x1);
    std::string name2= rSpecGen.StateName(x2);
    if(name2=="") name1=ToStringInteger(x2);
    std::string name12;
    if(m1requ) name12= name1 + "|" + name2 + "|r1m";
    else name12= name1 + "|" + name2 + "|r2m";
    name12=pResGen->UniqueStateName(name12);
    pResGen->StateName(x12,name12);
  }

#ifdef FAUDES_DEBUG_FUNCTION
  pResGen->DWrite();
  pResGen->Write("tmp_syn_xxx_"+pResGen->Name()+".gen");
  pResGen->WriteStateSet(specmarking);
  pResGen->WriteStateSet(plantmarking);
#endif

  // make resulting generator trim until it's fully controllable
  while (1) {
    if (pResGen->Empty()) {
      break;
    }
    Idx state_num = pResGen->Size();
    SupConUnchecked(rPlantGen, rCAlph, *pResGen);
    pResGen->Trim();
    pResGen->Complete();
    OmegaControlledLiveness(*pResGen,rCAlph,plantmarking,specmarking);
    if(pResGen->Size() == state_num) {
      break;
    }
  }

  // fix statenames ...
  //std::map< OPSState, Idx>::iterator rcit;
  if(rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && pResGen->StateNamesEnabled()) 
  for(rcit=rReverseCompositionMap.begin(); rcit!=rReverseCompositionMap.end(); rcit++) {
    Idx x1=rcit->first.q1;
    Idx x2=rcit->first.q2;
    bool m1requ=rcit->first.m1required;
    Idx x12=rcit->second;
    if(!pResGen->ExistsState(x12)) continue;
    std::string name1= rPlantGen.StateName(x1);
    if(name1=="") name1=ToStringInteger(x1);
    std::string name2= rSpecGen.StateName(x2);
    if(name2=="") name1=ToStringInteger(x2);
    std::string name12;
    if(m1requ) name12= name1 + "|" + name2 + "|r1m";
    else name12= name1 + "|" + name2 + "|r2m";
    name12=pResGen->UniqueStateName(name12);
    pResGen->StateName(x12,name12);
  }

  // .. or clear them (?)
  if(!(rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && pResGen->StateNamesEnabled())) 
    pResGen->StateNamesEnabled(false);

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}


/*
****************************************
* SUPCON: WRAPPER / USER FUNCTIONS     *
****************************************
*/


// OmegaSupConNB(rPlantGen, rCAlph, rSpecGen, rResGen)
void OmegaSupConNB(
  const Generator& rPlantGen, 
  const EventSet& rCAlph, 
  const Generator& rSpecGen, 
  Generator& rResGen) 
{

  // CONSISTENCY CHECK:
  ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSpecGen);

   
  // HELPERS:
  std::map< OPSState , Idx> rcmap;

  // ALGORITHM:
  OmegaSupConNBUnchecked(rPlantGen, rCAlph, rSpecGen, rcmap, rResGen);
}





// omega supcon for Systems:
// uses and maintains controllablity from plant 
void OmegaSupConNB(
  const System& rPlantGen, 
  const Generator& rSpecGen, 
  Generator& rResGen) {

  // prepare result
  Generator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }

  // execute 
  OmegaSupConNB(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,*pResGen);

  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}


} // name space 
